(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2007,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_base.ml,v 1.15 2007/04/26 16:18:32 blanqui Exp $ *)

(* Base definitions for generation functions *)

open Spec;;
open Parsetree;;
open Code;;
open Relation;;
open Genr_pattern;;
open Genr_sharing;;

(* Do we generate maximum sharing ? *)

let set_sharable_target, get_sharable_target =
  let sharable_target = ref false in
  (fun () -> sharable_target := true),
  (fun () -> !sharable_target);;

let get_eq_fun_def () =
  if get_sharable_target () then ["%eq"] else ["%equal"];;

let genr_eq tname =
  let tdecl = Check.find_tname tname in
  Pstr_primitive
   ("eq_" ^ tname,
    { pval_type = Otype.eq_type tname tdecl.ptype_params;
      pval_prim = get_eq_fun_def ();
    }
   );;

(* Generation of a pattern *)

let genr_lhs p =
  if get_sharable_target ()
  then add_underscores (remove_topconstr p)
  else remove_topconstr p;;

(* Generation function for Rule (p1, p2). *)

let genr_rule p1 eopt p2 =
  let lhs = genr_lhs p1
  and rhs = Genr_pattern.lower_case p2 in
  match eopt with
    | None -> Clause (Ocaml_pattern lhs, Ocaml_pattern rhs)
    | Some e ->	Clause_when
	(Ocaml_pattern lhs, Ocaml_expression e, Ocaml_pattern rhs);;

(* Getting the construction function name out of the constructor name. *)

let construction_function_name = String.lowercase;;

(* Name generation functions from a constructor name:
   [prefix_cname p c] is the name "prefix_name",
   where "prefix" is the contents of p,
   and "name" is the lower case version of c. *)
let prefix_cname prefix cname =
  Printf.sprintf "%s_%s" prefix (construction_function_name cname);;

let constr_fun1 cname = app1 (construction_function_name cname);;
let constr_fun2 cname x y = constr_fun1 cname (pair x y);;

(* Term construction. *)

let constr cname args =
   if get_sharable_target ()
   then App (Id (Genr_sharing.mk_constr cname), args)
   else Constr (cname, args);;

let constr1 cname x = constr cname [x];;
let constr2 cname x y = constr cname [x; y];;

(* Pattern construction. *)

let pattern cname args =
   if get_sharable_target ()
   then Genr_sharing.constr_underscore (cname, args)
   else Constr (cname, args);;

let pattern1 cname x = pattern cname [x];;
let pattern2 cname x y = pattern cname [x; y];;

(* Data structure for generation of a construction function. *)

type function_definition = {
  fd_uname : uident;
  fd_relations : Parsetree.relations;
  fd_arity : int;
};;
